SystemOrganization addCategory: #ControlTests!

TestCase subclass: #ControlTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ControlTests'!

!ControlTests commentStamp: 'fbs 3/31/2011 17:34' prior: 0!
This class exercises the shift/reset control operators.!

!ControlTests methodsFor: 'testing' stamp: 'fbs 4/15/2011 16:42'!
testContinuationsCompose
	| add1 add2 toString |
	[ 1 + [ :k |
			add1 := k.
			k value: 0 ] shift ] reset.

	[ 2 + [ :k |
		add2 := k.
		k value: 0 ] shift ] reset.

	self assert: (add1 value: (add2 value: 0)) = 3.
	
	"shift -> [ :x | x printString ]"
	[ [ :k | toString := k. k value: 0 ] shift printString ] reset.

	self assert: (toString value: (add1 value: 0)) = '1'.! !

!ControlTests methodsFor: 'testing - exceptions' stamp: 'fbs 4/14/2011 17:24'!
testExceptionWithinShiftEscapesContinuation
	| fun returnValue |
	fun := [ Exception signal: [ :k | k value: '2' ] shift ].
	self
		should: [ fun reset ]
		raise: Exception
		description: 'Reset ate the thrown exception'.
		
	[ fun reset ]
		on: Exception do: [ :e | returnValue := e messageText ].
	self assert: returnValue = '2' description: 'k returned wrong value'.! !

!ControlTests methodsFor: 'testing' stamp: 'fbs 4/14/2011 17:37'!
testNestedResetsDoNothing
	| fun |
	fun := [ [ 2 ] reset ].
	self assert: fun value = fun reset.
	self assert: fun reset = 2.! !

!ControlTests methodsFor: 'testing' stamp: 'fbs 4/14/2011 17:51'!
testNestedShiftResetsNest
	| innerCont outerCont |
	[1 + [ :k |
		outerCont := k. "stack reified as [ :x | 1 + x ]"
		k value: [ 2 + [ :kk |
				innerCont := kk. "stack reified as [ :x | 2 + x ]"
				kk value: 0 ] shift ] reset ] shift ] reset.

	self assert: (innerCont value: 1) = 3.
	self assert: (outerCont value: 1) = 2.! !

!ControlTests methodsFor: 'testing - unwinds' stamp: 'fbs 4/18/2011 22:28'!
testRepeatedInvocationsShouldAlwaysRunEnsures
	| cont count |
	"[:x | 1 + ( [ x ] ensure: [ count := count + 1 ] ) ]"
	count := 0.
	[ 2 + ( [ [ :k | cont := k. k value: 0] shift ] ensure: [ count := count + 1 ] ) ] reset.
	self assert: count = 1 description: 'Ensure didn''t run the first time'.
	
	self assert: (cont value: 1) = 3.
	self assert: count = 2 description: 'Ensure didn''t run on 2nd invocation'.
	
	self assert: (cont value: 3) = 5.
	self assert: count = 3 description: 'Ensure didn''t run on 3rd invocation'.! !

!ControlTests methodsFor: 'testing' stamp: 'fbs 3/31/2011 17:36'!
testResetDoesNothingOnItsOwn
	| fun |
	fun := [1 + 2].
	self assert: fun value = fun reset.! !

!ControlTests methodsFor: 'testing - exceptions' stamp: 'fbs 4/14/2011 16:41'!
testResetDoesntEatExceptions
	self
		should: [ [ Exception signal ] reset ]
		raise: Exception
		description: 'Reset ate the thrown exception'.! !

!ControlTests methodsFor: 'testing' stamp: 'fbs 4/14/2011 17:51'!
testShiftCutsToNearestReset
	| cont |
	[1 + [ 2 + [ :k |
			cont := k.
			k value: 0 ] shift ] reset ] reset.
	self assert: (cont value: 0) = 2.! !

!ControlTests methodsFor: 'testing' stamp: 'fbs 4/14/2011 17:27'!
testShiftMakesReentrantContinuation
	| cont |
	self assert: [ 2 + [ :k | cont := k. k value: 0 ] shift ] reset = 2.
	
	self assert: (cont value: 2) = 4.
	self assert: (cont value: 4) = 6.! !

!ControlTests methodsFor: 'testing' stamp: 'fbs 4/14/2011 17:57'!
testShiftReifiesStackAsFunction
	"Stack reified as [:x | 2 + x] so the expression reduces to
		(1 + (3 + [:x | 2 + x] value: 4))
		== (1 + (3 + (2 + 4)))
		== 10."
	self assert: (1 + [ 2 + [ :k | 3 + (k value: 4) ] shift ] reset) = 10.
	
	"Stack again reified as [:x | 2 + x] so the expression reduces to
		(1 + (3 + [:x | 2 + x] value: 5 + [:x | 2 + x] value: 1))
		== (1 + (3 + (2 + 5) + (2 + 1)))
		== 14."
	self assert: (1 + [ 2 + [ :k | 3 + (k value: 5) + (k value: 1) ] shift ] reset) = 14.! !

!ControlTests methodsFor: 'testing - exceptions' stamp: 'fbs 4/15/2011 13:51'!
testShiftWithCaughtException
	| fun |
	fun := [1 + [:k | k value: ([ 2 / 0 ] on: ZeroDivide do: [ 1 ]) ] shift ].
	self shouldnt: [fun reset] raise: ZeroDivide.! !

!ControlTests methodsFor: 'testing - unwinds' stamp: 'fbs 4/18/2011 22:12'!
testShiftWithExceptionAndInnerEnsureRaisesExceptionAndRunsEnsure
	| flag fun |
	flag := false.
	fun := [ 2 + [ :k | [ 1 / 0 ] ensure: [ flag := true ] ] shift ].
	self should: [ fun reset ] raise: ZeroDivide.
	self assert: flag description: 'Ensure block didn''t run'.! !

!ControlTests methodsFor: 'testing - unwinds' stamp: 'fbs 4/15/2011 13:11'!
testShiftWithInnerEnsure
	| flag |
	flag := false.
	[ 2 + [ :k | [ 0 ] ensure: [ flag := true ] ] shift ] reset.
	self assert: flag description: 'Ensure block didn''t run'.! !

!ControlTests methodsFor: 'testing' stamp: 'fbs 4/14/2011 15:20'!
testShiftWithNoResetRaisesException
	self
		should: [[ :k | 1 + 2] shift]
		raise: MissingResetException
		description: 'No missing reset exception for BlockClosure>>shift'.! !

!ControlTests methodsFor: 'testing - unwinds' stamp: 'fbs 4/15/2011 13:47'!
testShiftWithOuterEnsure
	| cont flag |

	[ [ 2 + [ :k |
			cont := k.
			k value: 0 ] shift ] reset ] ensure: [ flag := true ].
	self assert: flag description: 'Ensure block didn''t run'.
	
	flag := false.
	cont value: 0.
	self deny: flag description: 'Ensure block ran again!!'.! !

!ControlTests methodsFor: 'testing' stamp: 'fbs 3/31/2011 17:54'!
testUnusedShiftThrowsAwayCallStack
	self assert: 1 + [2 + [:k | 3] shift] reset = 4.! !
